from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from decimal import ROUND_DOWN, Decimal, localcontext
from pathlib import Path
from sys import platform
from typing import Callable, Literal

import numpy as np
import pandas as pd
import requests

from ._ledger import (
    Account,
    Currency,
    GeneralLedgerLine,
    GeneralLedgerLineImpl,
    LedgerElement,
    LedgerElementImpl,
)


def _get_currency(
    lines: Sequence[GeneralLedgerLine[Account, Currency]],
) -> Sequence[Currency]:
    return np.unique([x.currency for line in lines for x in line.values])


def get_prices(
    currency: Iterable[Currency],
) -> Mapping[Currency, "pd.Series[Decimal]"]:
    """
    Get prices of the currency indexed by date (not datetime).

    Interpolates missing dates by forward fill.

    Parameters
    ----------
    currency : Iterable[Currency]
        The currency to get prices.

    Returns
    -------
    Mapping[Currency, pd.Series[Decimal]]
        The prices, indexed by date.

    """
    URL = "https://www.mizuhobank.co.jp/market/quote.csv"
    path_dir = Path("~/.cache/aoiro").expanduser()
    path_dir.mkdir(exist_ok=True, parents=True)
    path = path_dir / "quote.csv"
    if not path.exists():
        with path.open("w", encoding="utf-8") as f:
            r = requests.get(URL, timeout=5)
            r.encoding = "Shift_JIS"
            f.write(r.text)
    df = pd.read_csv(
        path,
        index_col=0,
        skiprows=3 if platform == "win32" else 2,
        na_values=["*****"],
        parse_dates=True,
    )
    # fill missing dates
    df = df.reindex(pd.date_range(df.index[0], df.index[-1]), method="ffill")
    return df


def multidimensional_ledger_to_ledger(
    lines: Sequence[GeneralLedgerLine[Account, Currency]],
    is_debit: Callable[[Account], bool],
    prices: Mapping[Currency, "pd.Series[Decimal]"] = {},
) -> Sequence[
    GeneralLedgerLineImpl[Account | Literal["為替差益", "為替差損"], Literal[""]]
]:
    """
    Convert multidimensional ledger to ledger.

    The multidimensional ledger's any account must have
    positive amount.
    At the BOY, the previous B/S sheet
    must be added as a general ledger line, for example,

    >>> lines.append(
    >>>    GeneralLedgerLineImpl(
    >>>        date=datetime(2024, 1, 1),
    >>>        values=[
    >>>            ("売掛金", 1000, ""),
    >>>            ("元入金", 1000, "")
    >>>         ]
    >>>     )
    >>> )

    Returns
    -------
    GeneralLedgerLineImpl[Account | Literal["為替差益", "為替差損"], Literal[""]]
        The ledger lines.

    """
    # get prices, use prices_
    lines = sorted(lines, key=lambda x: x.date)
    prices_ = dict(prices)
    del prices
    prices_.update(get_prices(set(_get_currency(lines)) - set(prices_.keys())))

    # balance of (account, currency) represented by tuple (price, amount)
    balance: dict[Account, dict[Currency, list[tuple[Decimal, Decimal]]]] = defaultdict(
        lambda: defaultdict(list)
    )

    # the new lines
    lines_new = []
    for line in lines:
        # profit of the old line
        profit = Decimal(0)

        # the values of the new line
        values: list[
            LedgerElement[Account | Literal["為替差益", "為替差損"], Literal[""]]
        ] = []

        # iterate over the values of the old line
        for el in line.values:
            amount_left = el.amount
            # skip if the currency is empty (default, exchange rate = 1)
            if el.currency == "":
                values.append(el)  # type: ignore
                continue

            # meaning of "quote": https://docs.ccxt.com/#/?id=market-structure
            price_current = Decimal(str(prices_[el.currency][line.date]))
            amount_in_quote = Decimal(0)
            if amount_left < 0:
                while amount_left < 0:
                    # the maximum amount to subtract from the first element
                    # of the balance[account][currency]
                    amount_substract = max(
                        -amount_left, balance[el.account][el.currency][0][1]
                    )
                    amount_in_quote -= (
                        amount_substract * balance[el.account][el.currency][0][0]
                    )
                    amount_left += amount_substract

                    # subtract the amount from the balance
                    balance[el.account][el.currency][0] = (
                        balance[el.account][el.currency][0][0],
                        balance[el.account][el.currency][0][1] - amount_substract,
                    )
                    # remove if the amount is zero
                    if balance[el.account][el.currency][0][1] == 0:
                        balance[el.account][el.currency].pop(0)
            else:
                balance[el.account][el.currency].append((price_current, amount_left))
                amount_in_quote = amount_left * price_current

            # round the amount_in_quote
            with localcontext() as ctx:
                ctx.rounding = ROUND_DOWN
                amount_in_quote = round(amount_in_quote, 0)

            # append to the values of the new line
            values.append(
                LedgerElementImpl(
                    account=el.account, amount=amount_in_quote, currency=""
                )
            )

            # add to the profit
            if is_debit(el.account) is True:
                profit += amount_in_quote
            else:
                profit -= amount_in_quote

        # append only if profit is not zero
        if profit != 0:
            values.append(
                LedgerElementImpl(
                    account="為替差益" if profit > 0 else "為替差損",
                    amount=abs(profit),
                    currency="",
                )
            )

        # append the new line
        lines_new.append(GeneralLedgerLineImpl(date=line.date, values=values))
    return lines_new



import re
import warnings
from collections.abc import Iterable
from decimal import Decimal
from pathlib import Path
from typing import Any

import pandas as pd
from dateparser import parse

from .._ledger import GeneralLedgerLineImpl, LedgerElementImpl


def read_all_csvs(path: Path, /, **kwargs: Any) -> pd.DataFrame:
    """
    Read all CSV files in the path.

    Parameters
    ----------
    path : Path
        The path to the directory containing CSV files.
    **kwargs : Any
        The keyword arguments for `pd.read_csv`.

    Returns
    -------
    pd.DataFrame
        The concatenated DataFrame with
        column "path" containing the relative path of the CSV file added.

    """
    dfs = []
    for p in path.rglob("*.csv"):
        df = pd.read_csv(p, **kwargs)
        df["path"] = p.relative_to(path).as_posix()
        dfs.append(df)
    if not dfs:
        return pd.DataFrame(columns=["path"])
    return pd.concat(dfs)


def parse_date(s: str) -> pd.Timestamp:
    """
    Parse date.

    Prefer the last day of the month if the day is not provided.

    Parameters
    ----------
    s : str
        The string to parse.

    Returns
    -------
    pd.Timestamp
        The parsed date.

    """
    return pd.Timestamp(parse(s, settings={"PREFER_DAY_OF_MONTH": "last"}))


def parse_money(
    s: str, currency: str | None = None
) -> tuple[Decimal | None, str | None]:
    """
    Parse money.

    Parameters
    ----------
    s : str
        The string to parse.
    currency : str | None, optional
        The currency, by default None.
        If provided, the currency
        in the string would be ignored and replaced by this.

    Returns
    -------
    tuple[Decimal | None, str | None]
        The amount and the currency.

    """
    match = re.search(r"[\d.]+", s)
    if match is None:
        return None, None
    amount = Decimal(match.group())
    if currency is None:
        currency = re.sub(r"\s+", "", s[: match.start()] + s[match.end() :])
    return amount, currency


def read_simple_csvs(path: Path) -> pd.DataFrame:
    """
    Read all CSV files in the path.

    The CSV files are assumed to have columns
    ["発生日", "金額"].

    Parameters
    ----------
    path : Path
        The path to the directory containing CSV files.

    Returns
    -------
    pd.DataFrame
        The concatenated DataFrame with columns
        ["発生日", "金額", "通貨", "path"].

    """
    df = read_all_csvs(path, dtype=str)
    for col in ["発生日", "金額"]:
        if col not in df.columns:
            df[col] = None

    # parse date
    for k in df.columns:
        if "日" not in k:
            continue
        df[k] = df[k].map(parse_date)

    # parse money
    df[["金額", "通貨"]] = pd.DataFrame(
        df["金額"].map(parse_money).tolist(), index=df.index
    )

    # set date as index
    df.set_index("発生日", inplace=True, drop=False)
    return df


def read_general_ledger(path: Path) -> Iterable[GeneralLedgerLineImpl[Any, Any]]:
    """
    Read general ledger.

    The first column is assumed to be the date.
    For all n in N. the 2n-1-th column is assumed to be
    the account name, and the 2n-th column
    is assumed to be the amount.

    Parameters
    ----------
    path : Path
        The path to the CSV file.

    Returns
    -------
    Iterable[GeneralLedgerLineImpl[Any, Any]]
        The general ledger.

    """
    df = read_all_csvs(path / "general", header=None, dtype=str)
    df.drop(columns="path", inplace=True)
    if df.empty:
        return
    if len(df.columns) % 2 != 1:
        raise ValueError("The number of columns should be odd.")
    if len(df.columns) < 3:
        raise ValueError("The number of columns should be at least 3.")
    for _, row in df.iterrows():
        values: list[LedgerElementImpl[Any, Any]] = []
        for i in range(1, len(row), 2):
            amount, currency = parse_money(row[i + 1])
            if amount is None:
                warnings.warn(f"Amount not found in {row[i + 1]}", stacklevel=2)
                continue
            values.append(
                LedgerElementImpl(
                    account=row[i],
                    amount=amount,
                    currency=currency,
                )
            )
        yield GeneralLedgerLineImpl(
            values=values,
            date=parse_date(row[0]),
        )
